from .config import PolygonConfig

import datetime
import os
import pandas as pd
import polygon
import logging
from concurrent.futures import ProcessPoolExecutor


def configure_logging():
    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s %(processName)s %(levelname)s %(message)s",
        handlers=[logging.StreamHandler()],
    )


def fetch_and_save_tickers_for_date(config: PolygonConfig, date: pd.Timestamp):
    try:
        configure_logging()
        logger = logging.getLogger()
        logger.info(f"Fetching tickers for {date}")
        assets = PolygonAssets(config)
        all_tickers = assets.fetch_all_tickers(date=date)
        assets.save_tickers_for_date(all_tickers, date)
    except Exception as e:
        logger.exception(f"Error fetching tickers for {date}: {e}")


def unique_simple_list(x):
    return [str(y) for y in x.unique() if pd.notnull(y)]


def unique_simple_date_list(x):
    return [y.date().isoformat() for y in x.unique() if pd.notnull(y)]


class PolygonAssets:
    def __init__(self, config: PolygonConfig):
        self.config = config
        self.polygon_client = polygon.RESTClient(api_key=config.api_key)

    def fetch_tickers(
        self,
        date: pd.Timestamp,
        active: bool = True,
    ):
        response = self.polygon_client.list_tickers(
            market=self.config.market, active=active, date=date.date(), limit=500
        )
        tickers_df = pd.DataFrame(list(response))
        # The currency info is for crypto.  The source_feed is always NA.
        tickers_df.drop(
            columns=[
                "currency_symbol",
                "base_currency_symbol",
                "base_currency_name",
                "source_feed",
            ],
            inplace=True,
        )

        tickers_df["request_date"] = date
        tickers_df["last_updated_utc"] = pd.to_datetime(tickers_df["last_updated_utc"])
        tickers_df["delisted_utc"] = pd.to_datetime(tickers_df["delisted_utc"])

        # Make sure there are no leading or trailing spaces in the column values
        for col in tickers_df.columns:
            if tickers_df[col].dtype == "string":
                # Still gotta check value types because of NA values.
                tickers_df[col] = tickers_df[col].apply(
                    lambda x: x.strip() if isinstance(x, str) else x
                )

        # Test tickers have no CIK value.
        # Actually sometimes listed tickers have no CIK value.
        # Case in point is ALTL ETF which doesn't include a CIK in the Ticker search response for 2024-07-01
        # but did on 2024-06-25.  Suspiciouly 4 years after the listing date of 2020-06-25.
        # We'll have to leave this cleaning of test tickers until we can call Ticker Details.
        # tickers_df = tickers_df.dropna(subset=["ticker", "cik"])

        return tickers_df

    def validate_active_tickers(self, tickers: pd.DataFrame):
        # # All tickers are active
        # inactive_tickers = (
        #     tickers[~tickers.index.get_level_values("active")]["ticker"].unique().tolist()
        # )
        # assert (
        #     not inactive_tickers
        # ), f"{len(inactive_tickers)} tickers are not active: {inactive_tickers[:15]}"

        # No tickers with missing last_updated_utc
        missing_last_updated_utc_tickers = (
            tickers[tickers["last_updated_utc"].isnull()]["ticker"].unique().tolist()
        )
        assert (
            not missing_last_updated_utc_tickers
        ), f"{len(missing_last_updated_utc_tickers)} tickers have missing last_updated_utc: {missing_last_updated_utc_tickers[:15]}"

        # # No tickers with missing name
        # missing_name_tickers = (
        #     tickers[tickers["name"].isnull()]["ticker"].unique().tolist()
        # )
        # if missing_name_tickers:
        #     logging.warning(
        #         f"{len(missing_name_tickers)} tickers have missing name: {missing_name_tickers[:15]}"
        #     )
        # assert (
        #     not missing_name_tickers
        # ), f"{len(missing_name_tickers)} tickers have missing name: {missing_name_tickers[:15]}"

        # No tickers with missing locale
        missing_locale_tickers = (
            tickers[tickers["locale"].isnull()]["ticker"].unique().tolist()
        )
        assert (
            not missing_locale_tickers
        ), f"{len(missing_locale_tickers)} tickers have missing locale: {missing_locale_tickers[:15]}"

        # No tickers with missing market
        missing_market_tickers = (
            tickers[tickers["market"].isnull()]["ticker"].unique().tolist()
        )
        assert (
            not missing_market_tickers
        ), f"{len(missing_market_tickers)} tickers have missing market: {missing_market_tickers[:15]}"

        # # No tickers with missing primary exchange
        # missing_primary_exchange_tickers = (
        #     tickers[tickers["primary_exchange"].isnull()]
        #     .index.get_level_values("ticker")
        #     .tolist()
        # )
        # # We'll just warn here and filter in the merge_tickers function.
        # if missing_primary_exchange_tickers:
        #     logging.warning(
        #         f"{len(missing_primary_exchange_tickers)} tickers have missing primary exchange: {missing_primary_exchange_tickers[:15]}"
        #     )
        # assert (
        #     not missing_primary_exchange_tickers
        # ), f"{len(missing_primary_exchange_tickers)} tickers have missing primary exchange: {missing_primary_exchange_tickers[:15]}"

        # # No tickers with missing type
        # missing_type_tickers = (
        #     tickers[tickers["type"].isnull()].index.get_level_values("ticker").tolist()
        # )
        # # assert not missing_type_tickers, f"{len(missing_type_tickers)} tickers have missing type: {missing_type_tickers[:15]}"
        # # Just a warning because there are legit tickers with missing type
        # if missing_type_tickers:
        #     logging.warning(
        #         f"{len(missing_type_tickers)} tickers have missing type: {missing_type_tickers[:15]}"
        #     )

        # No tickers with missing currency
        missing_currency_tickers = (
            tickers[tickers["currency_name"].isnull()]["ticker"].unique().tolist()
        )
        assert (
            not missing_currency_tickers
        ), f"{len(missing_currency_tickers)} tickers have missing currency_name: {missing_currency_tickers[:15]}"

    def fetch_all_tickers(self, date: pd.Timestamp):
        all_tickers = pd.concat(
            [
                self.fetch_tickers(date=date, active=True),
                self.fetch_tickers(date=date, active=False),
            ]
        )

        # We're keeping these tickers with missing type because it is some Polygon bug.
        # # Drop rows with no type.  Not sure why this happens but we'll ignore them for now.
        # active_tickers = active_tickers.dropna(subset=["type"])

        self.validate_active_tickers(all_tickers)

        return all_tickers

    def ticker_file_exists(self, date: pd.Timestamp):
        return os.path.exists(self.config.ticker_file_path(date))

    def save_tickers_for_date(self, tickers: pd.DataFrame, date: pd.Timestamp):
        tickers.to_parquet(self.config.ticker_file_path(date))

    def load_tickers_for_date(self, date: pd.Timestamp):
        try:
            tickers = pd.read_parquet(self.config.ticker_file_path(date))
            return tickers
        except (FileNotFoundError, IOError) as e:
            logging.error(f"Error loading tickers for {date}: {e}")
            try:
                # Remove the file so that it can be fetched again
                os.remove(self.config.ticker_file_path(date))
            except (FileNotFoundError, IOError) as e2:
                logging.error(
                    f"Error removing file {self.config.ticker_file_path(date)}: {e2}"
                )
            return None

    def load_all_tickers(self, fetch_missing=False, max_workers=None):
        dates = list(
            self.config.calendar.trading_index(
                start=self.config.start_timestamp,
                end=self.config.end_timestamp,
                period="1D",
            )
        )
        if fetch_missing:
            missing_dates = [
                date for date in dates if not self.ticker_file_exists(date)
            ]
            print(f"{len(missing_dates)=}")
            if missing_dates:
                max_workers = (
                    max_workers if max_workers is not None else self.config.max_workers
                )
                if max_workers == 0:
                    for date in missing_dates:
                        fetch_and_save_tickers_for_date(self.config, date)
                else:
                    print("Starting process pool executor")
                    with ProcessPoolExecutor(max_workers=max_workers) as executor:
                        executor.map(
                            fetch_and_save_tickers_for_date,
                            [self.config] * len(missing_dates),
                            missing_dates,
                        )
        dates_with_files = [date for date in dates if self.ticker_file_exists(date)]
        if fetch_missing and (len(dates_with_files) != len(dates)):
            logging.warning(
                f"Only {len(dates_with_files)} of {len(dates)} dates have files."
            )
        all_tickers = pd.concat(
            [self.load_tickers_for_date(date) for date in dates_with_files]
        )
        return all_tickers

    def merge_tickers(self, all_tickers: pd.DataFrame):
        all_tickers.set_index(
            ["ticker", "primary_exchange", "cik", "type", "composite_figi", "active"],
            drop=False,
            inplace=True,
        )
        all_tickers.sort_index(inplace=True)

        # Make sure there are no leading or trailing spaces in the column values
        # This also does some kind of change to the type for the StringArray values which makes Arrow Parquet happy.
        all_tickers = all_tickers.map(lambda x: x.strip() if isinstance(x, str) else x)

        # # Drops rows with missing primary exchange (rare but means it isn't actually active).
        # all_tickers.dropna(subset=["name", "primary_exchange"], inplace=True)

        # # Only keep rows with type CS, ETF, or ADRC
        # all_tickers = all_tickers[all_tickers["type"].isin(["CS", "ETF", "ADRC"])]

        merged_tickers = (
            all_tickers.groupby(
                level=[
                    "ticker",
                    "primary_exchange",
                    "cik",
                    "type",
                    "composite_figi",
                    "active",
                ],
                dropna=False,
            )
            .agg(
                {
                    "request_date": ["min", "max"],
                    "last_updated_utc": lambda x: x.max().date(),
                    "name": "unique",
                    "share_class_figi": unique_simple_list,
                    "delisted_utc": unique_simple_date_list,
                    "currency_name": "unique",
                    "locale": "unique",
                    "market": "unique",
                }
            )
            .sort_values(by=("request_date", "max"))
        )

        # Flatten the multi-level column index
        merged_tickers.columns = [
            "_".join(col).strip() for col in merged_tickers.columns.values
        ]

        # Rename the columns
        merged_tickers.rename(
            columns={
                "request_date_min": "start_date",
                "request_date_max": "end_date",
                "last_updated_utc_<lambda>": "last_updated_utc",
                "share_class_figi_unique_simple_list": "share_class_figi",
                "delisted_utc_unique_simple_date_list": "delisted_utc",
            },
            inplace=True,
        )
        merged_tickers.rename(
            columns=lambda x: x.removesuffix("_unique"),
            inplace=True,
        )

        all_tickers.sort_index(inplace=True)
        return merged_tickers


def list_to_string(x):
    if not hasattr(x, "__len__"):
        return str(x)
    if len(x) == 0:
        return ""
    if len(x) == 1:
        return str(x[0])
    s = set([str(y) for y in x])
    return f"[{']['.join(sorted(list(s)))}]"


def get_ticker_universe(config: PolygonConfig, fetch_missing: bool = False):
    tickers_csv_path = config.tickers_csv_path
    print(f"{tickers_csv_path=}")
    parquet_path = tickers_csv_path.removesuffix(".csv") + ".parquet"
    if not os.path.exists(parquet_path):
        if os.path.exists(tickers_csv_path):
            os.remove(tickers_csv_path)
        assets = PolygonAssets(config)
        all_tickers = assets.load_all_tickers(fetch_missing=fetch_missing)
        all_tickers.info()
        # all_tickers.to_csv(tickers_csv_path)
        logging.info("Merging tickers")
        merged_tickers = assets.merge_tickers(all_tickers)
        merged_tickers.info()
        merged_tickers.to_parquet(tickers_csv_path.removesuffix(".csv") + ".parquet")
        print(
            f"Saved {len(merged_tickers)} tickers to {tickers_csv_path.removesuffix('.csv') + '.parquet'}"
        )
    if not os.path.exists(tickers_csv_path):
        merged_tickers = pd.read_parquet(parquet_path)
        merged_tickers["name"] = merged_tickers["name"].apply(list_to_string)
        merged_tickers["share_class_figi"] = merged_tickers["share_class_figi"].apply(
            list_to_string
        )
        merged_tickers["delisted_utc"] = merged_tickers["delisted_utc"].apply(
            list_to_string
        )
        merged_tickers["currency_name"] = merged_tickers["currency_name"].apply(
            list_to_string
        )
        merged_tickers["locale"] = merged_tickers["locale"].apply(list_to_string)
        merged_tickers["market"] = merged_tickers["market"].apply(list_to_string)
        merged_tickers.to_csv(
            tickers_csv_path, escapechar="\\", quoting=csv.QUOTE_NONNUMERIC
        )
        print(f"Saved {len(merged_tickers)} tickers to {tickers_csv_path}")

    # merged_tickers = pd.read_csv(
    #     tickers_csv_path,
    #     escapechar="\\",
    #     quoting=csv.QUOTE_NONNUMERIC,
    #     dtype={
    #         "ticker": str,
    #         "primary_exchange": str,
    #         "cik": str,
    #         "type": str,
    #         "share_class_figi": str,
    #     },
    #     # converters={
    #     #     "ticker": lambda x: str(x),
    #     #     "start_date": lambda x: pd.to_datetime(x),
    #     #     "cik": lambda x: str(x) if x else None,
    #     #     "name": lambda x: str(x),
    #     #     "end_date": lambda x: pd.to_datetime(x),
    #     #     "composite_figi": lambda x: str(x).upper(),
    #     #     "share_class_figi": lambda x: str(x).upper(),
    #     #     "currency_name": lambda x: str(x).lower(),
    #     #     "locale": lambda x: str(x).lower(),
    #     #     "market": lambda x: str(x).lower(),
    #     #     "primary_exchange": lambda x: str(x).strip().upper(),
    #     #     "type": lambda x: str(x).upper(),
    #     # },
    # )
    merged_tickers = pd.read_parquet(parquet_path)
    merged_tickers.info()
    return merged_tickers


# Initialize ticker files in __main__.  Use CLI args to specify start and end dates.
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Initialize ticker files.")
    parser.add_argument(
        "--start-date",
        type=str,
        help="Start date in ISO format (YYYY-MM-DD)",
        default="2014-05-01",
    )
    parser.add_argument(
        "--end-date",
        type=str,
        help="End date in ISO format (YYYY-MM-DD)",
        default="2024-04-01",
    )
    args = parser.parse_args()

    start_date = (
        datetime.datetime.strptime(args.start_date, "%Y-%m-%d").date()
        if args.start_date
        else datetime.date.today()
    )
    end_date = (
        datetime.datetime.strptime(args.end_date, "%Y-%m-%d").date()
        if args.end_date
        else datetime.date.today()
    )

    all_tickers = load_all_tickers(start_date, end_date, fetch_missing=True)
    merged_tickers = merge_tickers(all_tickers)
    merged_tickers.to_csv(f"data/tickers/us_tickers_{start_date}-{end_date}.csv")
    ticker_names = ticker_names_from_merged_tickers(merged_tickers)
    print(ticker_names)



from .bundle import (
    register_polygon_equities_bundle,
    symbol_to_upper,
    polygon_equities_bundle_day,
    polygon_equities_bundle_minute,
)

from .config import PolygonConfig
from .concat_all_aggs import concat_all_aggs_from_csv, generate_csv_agg_tables
from .adjustments import load_splits, load_dividends

__all__ = [
    "register_polygon_equities_bundle",
    "symbol_to_upper",
    "polygon_equities_bundle_day",
    "polygon_equities_bundle_minute",
    "PolygonConfig",
    "concat_all_aggs_from_csv",
    "generate_csv_agg_tables",
    "load_splits",
    "load_dividends",
]



from zipline.data.bundles import register
from zipline.data.resample import minute_frame_to_session_frame

from .config import PolygonConfig
from .concat_all_aggs import concat_all_aggs_from_csv, generate_csv_agg_tables
from .adjustments import load_splits, load_dividends

import pyarrow
import pyarrow.compute

import pandas as pd
import logging

import concurrent.futures


# TODO: Change warnings to be relative to number of days in the range.


def symbol_to_upper(s: str) -> str:
    if s.isupper():
        return s
    return "".join(map(lambda c: ("^" + c.upper()) if c.islower() else c, s))


def generate_all_agg_tables_from_csv(
    config: PolygonConfig,
):
    paths, schema, tables = generate_csv_agg_tables(config)
    for table in tables:
        table = table.sort_by([("ticker", "ascending"), ("window_start", "ascending")])
        yield table


# def remove_duplicated_index(df: pd.DataFrame) -> pd.DataFrame:
#     duplicated_index = df.index.duplicated(keep=False)
#     if not duplicated_index.any():
#         return df
#     # Find duplicate index values (date) with zero volume or transactions
#     duplicated_index_with_zero_activity = duplicated_index & (
#         df["volume"] == 0) | (df["transactions"] == 0)
#     if duplicated_index_with_zero_activity.any():
#         print(
#             f" WARNING: Got dupes with zero activity {df[duplicated_index_with_zero_activity]=}"
#         )
#         df = df[~duplicated_index_with_zero_activity]
#         duplicated_index = df.index.duplicated(keep=False)
#         if not duplicated_index.any():
#             return df
#     print(f" WARNING: Dropping dupes {df[duplicated_index]=}")
#     df = df[df.index.duplicated(keep="first")]
#     return df


def aggregate_multiple_aggs_per_date(df: pd.DataFrame) -> pd.DataFrame:
    duplicated_index = df.index.duplicated(keep=False)
    if not duplicated_index.any():
        return df
    duplicates = df[duplicated_index]
    duplicate_index_values = duplicates.index.values
    print()
    if duplicates["symbol"].nunique() != 1:
        logging.error(f"{duplicates['symbol'].unique()=} {duplicate_index_values=}")
    logging.warning(
        f"Aggregating dupes df[df.index.duplicated(keep=False)]=\n{duplicates}"
    )
    df = df.groupby(df.index).agg(
        {
            "symbol": "first",
            "volume": "sum",
            "open": "first",
            "close": "last",
            "high": "max",
            "low": "min",
            "transactions": "sum",
        }
    )
    print(f"WARNING: Aggregated dupes df=\n{df[df.index.isin(duplicate_index_values)]}")
    return df


def process_day_aggregates(
    table,
    sessions,
    metadata,
    calendar,
    symbol_to_sid: dict[str, int],
    dates_with_data: set,
):
    for symbol, sid in symbol_to_sid.items():
        df = table.filter(
            pyarrow.compute.field("symbol") == pyarrow.scalar(symbol)
        ).to_pandas()
        # The SQL schema zipline uses for symbols ignores case
        sql_symbol = symbol_to_upper(symbol)
        df["symbol"] = sql_symbol
        df["day"] = pd.to_datetime(df["day"].dt.date)
        df = df.set_index("day")
        if not df.index.is_monotonic_increasing:
            print(f" INFO: {symbol=} {sid=} not monotonic increasing")
            df.sort_index(inplace=True)
            # Remove duplicates
            df = df[~df.index.duplicated(keep="first")]
        # Take days as per calendar
        df = df[df.index.isin(sessions)]
        # 2019-08-13 has a bunch of tickers with multiple day aggs per date
        df = aggregate_multiple_aggs_per_date(df)
        if len(df) < 1:
            continue
        # Check first and last date.
        start_date = df.index[0]
        dates_with_data.add(start_date.date())
        end_date = df.index[-1]
        dates_with_data.add(end_date.date())
        try:
            duplicated_index = df.index.duplicated(keep=False)
            df_with_duplicates = df[duplicated_index]
            if len(df_with_duplicates) > 0:
                print(f" WARNING: {symbol=} {sid=} {len(df_with_duplicates)=}")
                df_with_duplicates.info()
                print(df_with_duplicates)
            # Synch to the official exchange calendar
            df = df.reindex(sessions.tz_localize(None))
        except ValueError as e:
            print(f" ERROR: {symbol=} {sid=} {e}")
            print(f"{start_date=} {end_date=} {sessions[0]=} {sessions[-1]=}")
            df.info()
        # Missing volume and transactions are zero
        df["volume"] = df["volume"].fillna(0)
        df["transactions"] = df["transactions"].fillna(0)
        # TODO: These fills should have the same price for OHLC (open for backfill, close for forward fill)
        # Forward fill missing price data (better than backfill)
        df.ffill(inplace=True)
        # Back fill missing data (maybe necessary for before the first day bar)
        df.bfill(inplace=True)
        # There should be no missing data
        if df.isnull().sum().sum() > 0:
            print(f" WARNING: Missing data for {symbol=} {sid=}")

        # The auto_close date is the day after the last trade.
        auto_close_date = end_date + pd.Timedelta(days=1)

        # Add a row to the metadata DataFrame. Don't forget to add an exchange field.
        metadata.loc[sid] = (
            start_date,
            end_date,
            auto_close_date,
            sql_symbol,
            calendar.name,
            symbol,
        )
        if len(df) > 0:
            yield sid, df
    return


def rename_polygon_to_zipline(table: pyarrow.Table, time_name: str) -> pyarrow.Table:
    table = table.rename_columns(
        [
            (
                "symbol"
                if name == "ticker"
                else time_name if name == "window_start" else name
            )
            for name in table.column_names
        ]
    )
    return table


def polygon_equities_bundle_day(
    environ,
    asset_db_writer,
    minute_bar_writer,
    daily_bar_writer,
    adjustment_writer,
    calendar,
    start_session,
    end_session,
    cache,
    show_progress,
    output_dir,
):
    config = PolygonConfig(
        environ=environ,
        calendar_name=calendar.name,
        start_session=start_session,
        end_session=end_session,
        agg_time="day",
    )

    by_ticker_aggs_arrow_dir = concat_all_aggs_from_csv(config)
    aggregates = pyarrow.dataset.dataset(by_ticker_aggs_arrow_dir)

    # Zipline uses case-insensitive symbols, so we need to convert them to uppercase with a ^ prefix when lowercase.
    # This is because the SQL schema zipline uses for symbols ignores case.
    # We put the original symbol in the asset_name field.
    metadata = pd.DataFrame(
        columns=(
            "start_date",
            "end_date",
            "auto_close_date",
            "symbol",
            "exchange",
            "asset_name",
        )
    )

    table = aggregates.to_table()
    table = rename_polygon_to_zipline(table, "day")
    # Get all the symbols in the table by using value_counts to tabulate the unique values.
    # pyarrow.Table.column returns a pyarrow.ChunkedArray.
    # https://arrow.apache.org/docs/python/generated/pyarrow.ChunkedArray.html#pyarrow.ChunkedArray.value_counts
    symbols = sorted(table.column("symbol").value_counts().field(0).to_pylist())
    symbol_to_sid = {symbol: sid for sid, symbol in enumerate(symbols)}
    dates_with_data = set()

    # Get data for all stocks and write to Zipline
    daily_bar_writer.write(
        process_day_aggregates(
            table=table,
            sessions=calendar.sessions_in_range(start_session, end_session),
            metadata=metadata,
            calendar=calendar,
            symbol_to_sid=symbol_to_sid,
            dates_with_data=dates_with_data,
        ),
        show_progress=show_progress,
    )

    # Write the metadata
    asset_db_writer.write(equities=metadata)

    # Load splits and dividends
    first_start_end = min(dates_with_data)
    last_end_date = max(dates_with_data)
    splits = load_splits(config, first_start_end, last_end_date, symbol_to_sid)
    dividends = load_dividends(config, first_start_end, last_end_date, symbol_to_sid)

    # Write splits and dividends
    adjustment_writer.write(splits=splits, dividends=dividends)


def process_minute_fragment(
    fragment,
    sessions,
    minutes,
    metadata,
    calendar,
    symbol_to_sid: dict[str, int],
    dates_with_data: set,
    agg_time: str,
):
    table = fragment.to_table()
    print(f" {table.num_rows=}")
    table = rename_polygon_to_zipline(table, "timestamp")
    table = table.sort_by([("symbol", "ascending"), ("timestamp", "ascending")])
    table = table.filter(pyarrow.compute.field("timestamp").isin(minutes))
    table_df = table.to_pandas()
    for symbol, df in table_df.groupby("symbol"):
        # print(f"\n{symbol=} {len(df)=} {df['timestamp'].min()} {df['timestamp'].max()}")
        if symbol not in symbol_to_sid:
            symbol_to_sid[symbol] = len(symbol_to_sid) + 1
        sid = symbol_to_sid[symbol]
        # The SQL schema zipline uses for symbols ignores case
        sql_symbol = symbol_to_upper(symbol)
        df["symbol"] = sql_symbol
        df = df.set_index("timestamp")
        if agg_time == "day":
            df.drop(columns=["symbol", "transactions"], inplace=True)
            # Check first and last date.
            start_date = df.index[0].date()
            start_timestamp = df.index[0]
            dates_with_data.add(start_date)
            end_date = df.index[-1].date()
            end_timestamp = df.index[-1]
            dates_with_data.add(end_date)
            df = df[df.index.isin(minutes)]
            len_before = len(df)
            if len(df) < 1:
                # TODO: Move sid assignment until after this check for no data.
                print(
                    f" WARNING: No data for {symbol=} {sid=} {len_before=} {start_timestamp=} {end_timestamp=}"
                )
                continue
            df = minute_frame_to_session_frame(df, calendar)
            df["symbol"] = sql_symbol
            df = df[df.index.isin(sessions)]

            # The auto_close date is the day after the last trade.
            auto_close_date = end_date + pd.Timedelta(days=1)

            # If metadata already has this sid, just extend the end_date and ac_date.
            if sid in metadata.index:
                if metadata.loc[sid, "start_date"] >= start_date:
                    print(
                        f" ERROR: {symbol=} {sid=} {metadata.loc[sid, 'start_date']=} >= {start_date=}"
                    )
                if metadata.loc[sid, "end_date"] >= start_date:
                    print(
                        f" ERROR: {symbol=} {sid=} {metadata.loc[sid, 'end_date']=} >= {end_date=}"
                    )
                metadata.loc[sid, "end_date"] = end_date
                metadata.loc[sid, "auto_close_date"] = auto_close_date
            else:
                # Add a row to the metadata DataFrame. Don't forget to add an exchange field.
                metadata.loc[sid] = (
                    start_date,
                    end_date,
                    auto_close_date,
                    symbol_to_upper(symbol),
                    calendar.name,
                    symbol,
                )
            df = df.reindex(sessions.tz_localize(None))
            # df = df.reindex(sessions)
            # Missing volume and transactions are zero
            df["volume"] = df["volume"].fillna(0)
            # df["transactions"] = df["transactions"].fillna(0)
            # Forward fill missing price data (better than backfill)
            # TODO: These fills should have the same price for OHLC (open for backfill, close for forward fill)
            df.ffill(inplace=True)
            # Back fill missing data (maybe necessary for before the first day bar)
            df.bfill(inplace=True)
            if len(df) > 0:
                # print(f"\n{symbol=} {sid=} {len_before=} {start_timestamp=} {end_date=} {end_timestamp=} {len(df)=}")
                yield sid, df
            else:
                print(
                    f" WARNING: No day bars for {symbol=} {sid=} {len_before=} {start_date=} {start_timestamp=} {end_date=} {end_timestamp=}"
                )
        else:
            len_before = len(df)
            df = df[df.index.isin(minutes)]
            if len(df) < 2:
                print(
                    f" WARNING: Not enough data for {symbol=} {sid=} {len(df)=} {len_before=}"
                )
                continue

            # A df with 1 bar crashes zipline/data/bcolz_minute_bars.py", line 747
            # pd.Timestamp(dts[0]), direction="previous"
            if len(df) > 1:
                yield sid, df
            else:
                print(
                    f" WARNING: Not enough minute bars for {symbol=} {sid=} {len(df)=}"
                )
    return


def process_minute_aggregates(
    fragments,
    sessions,
    minutes,
    metadata,
    calendar,
    symbol_to_sid: dict[str, int],
    dates_with_data: set,
    agg_time: str,
):
    # We want to do this by Hive partition at a time because each ticker will be complete.
    for fragment in fragments:
        yield from process_minute_fragment(
            fragment=fragment,
            sessions=sessions,
            minutes=minutes,
            metadata=metadata,
            calendar=calendar,
            symbol_to_sid=symbol_to_sid,
            dates_with_data=dates_with_data,
            agg_time=agg_time,
        )

    # This doesn't seem to be hardly any faster than the above, something with the GIL?
    # Also to use this we'd need to make sure the symbol_to_sid and dates_with_data are thread safe.
    # with concurrent.futures.ThreadPoolExecutor(max_workers=12) as executor:
    #     futures = [
    #         executor.submit(
    #             process_minute_fragment,
    #             fragment,
    #             sessions,
    #             minutes,
    #             metadata,
    #             calendar,
    #             symbol_to_sid,
    #             dates_with_data,
    #             agg_time,
    #         )
    #         for fragment in fragments
    #     ]
    #     for future in concurrent.futures.as_completed(futures):
    #         yield from future.result()


def polygon_equities_bundle_minute(
    environ,
    asset_db_writer,
    minute_bar_writer,
    daily_bar_writer,
    adjustment_writer,
    calendar,
    start_session,
    end_session,
    cache,
    show_progress,
    output_dir,
):
    config = PolygonConfig(
        environ=environ,
        calendar_name=calendar.name,
        start_session=start_session,
        end_session=end_session,
        agg_time="minute",
    )

    by_ticker_aggs_arrow_dir = concat_all_aggs_from_csv(config)
    aggregates = pyarrow.dataset.dataset(by_ticker_aggs_arrow_dir)
    print(f"{aggregates.schema=}")
    # 3.5 billion rows for 10 years of minute data.
    # print(f"{aggregates.count_rows()=}")
    # Can't sort the dataset because that reads it all into memory.
    # aggregates = aggregates.sort_by([("ticker", "ascending"), ("window_start", "ascending")])
    # print("Sorted")

    # Zipline uses case-insensitive symbols, so we need to convert them to uppercase with a ^ prefix when lowercase.
    # This is because the SQL schema zipline uses for symbols ignores case.
    # We put the original symbol in the asset_name field.
    metadata = pd.DataFrame(
        columns=(
            "start_date",
            "end_date",
            "auto_close_date",
            "symbol",
            "exchange",
            "asset_name",
        )
    )

    symbol_to_sid = {}
    dates_with_data = set()

    # Get data for all stocks and write to Zipline
    daily_bar_writer.write(
        process_minute_aggregates(
            fragments=aggregates.get_fragments(),
            sessions=calendar.sessions_in_range(start_session, end_session),
            minutes=calendar.sessions_minutes(start_session, end_session),
            metadata=metadata,
            calendar=calendar,
            symbol_to_sid=symbol_to_sid,
            dates_with_data=dates_with_data,
            agg_time="day",
        ),
        show_progress=show_progress,
    )

    # Get data for all stocks and write to Zipline
    minute_bar_writer.write(
        process_minute_aggregates(
            fragments=aggregates.get_fragments(),
            sessions=calendar.sessions_in_range(start_session, end_session),
            minutes=calendar.sessions_minutes(start_session, end_session),
            metadata=metadata,
            calendar=calendar,
            symbol_to_sid=symbol_to_sid,
            dates_with_data=dates_with_data,
            agg_time="minute",
        ),
        show_progress=show_progress,
    )

    # Write the metadata
    asset_db_writer.write(equities=metadata)

    # Load splits and dividends
    first_start_end = min(dates_with_data)
    last_end_date = max(dates_with_data)
    splits = load_splits(config, first_start_end, last_end_date, symbol_to_sid)
    dividends = load_dividends(config, first_start_end, last_end_date, symbol_to_sid)

    # Write splits and dividends
    adjustment_writer.write(splits=splits, dividends=dividends)


def register_polygon_equities_bundle(
    bundlename,
    start_session=None,
    end_session=None,
    calendar_name="XNYS",
    agg_time="day",
    # ticker_list=None,
    # watchlists=None,
    # include_asset_types=None,
):
    if agg_time not in ["day", "minute"]:
        raise ValueError(f"agg_time must be 'day' or 'minute', not '{agg_time}'")
    register(
        bundlename,
        (
            polygon_equities_bundle_minute
            if agg_time == "minute"
            else polygon_equities_bundle_day
        ),
        start_session=start_session,
        end_session=end_session,
        calendar_name=calendar_name,
        # minutes_per_day=390,
        # create_writers=True,
    )


# if __name__ == "__main__":
#     logging.basicConfig(level=logging.WARNING)
#     os.environ["POLYGON_MIRROR_DIR"] = "/Volumes/Oahu/Mirror/files.polygon.io"
#     os.environ["ZIPLINE_ROOT"] = "/Volumes/Oahu/Workspaces/zipline"
#     config = PolygonConfig(
#         environ=os.environ,
#         calendar_name="XNYS",
#         # start_session="2003-10-01",
#         # start_session="2018-01-01",
#         start_session="2023-01-01",
#         # end_session="2023-01-12",
#         end_session="2023-12-31",
#         # end_session="2024-06-30",
#     )
#     splits = load_polygon_splits(config)
#     splits.info()
#     print(splits.head())
#     dividends = load_polygon_dividends(config)
#     dividends.info()
#     print(dividends.head())
#     tickers = set(
#         splits["ticker"].unique().tolist() + dividends["ticker"].unique().tolist()
#     )
#     print(f"{len(tickers)=}")
#     ticker_to_sid = {ticker: sid for sid, ticker in enumerate(tickers)}
#     splits = load_splits(config, ticker_to_sid)
#     splits.info()
#     print(splits.head())
#     dividends = load_dividends(config, ticker_to_sid)
#     dividends.info()
#     print(dividends.head())



from .config import PolygonConfig

import os
import glob
import pandas as pd
from concurrent.futures import ProcessPoolExecutor
import fastparquet as fp
from pathlib import Path


def split_aggs_by_ticker(aggs_path: Path, by_ticker_dir: Path, extension: str):
    aggs_pf = fp.ParquetFile(aggs_path)
    df = aggs_pf.to_pandas()
    print(df.info())
    print(df.head())
    df.sort_values(["ticker", "window_start"], inplace=True)
    for ticker in df["ticker"].unique():
        ticker_df = df[df["ticker"] == ticker]
        ticker_path = by_ticker_dir / f"{ticker}{extension}"
        if os.path.exists(ticker_path):
            fp.write(ticker_path, ticker_df, has_nulls=False, write_index=False, fixed_text={"ticker": len(ticker) + 1})
        else:            
            fp.write(ticker_path, ticker_df, append=True, has_nulls=False, write_index=False)


def split_all_aggs_by_ticker(
    aggs_dir,
    by_ticker_dir,
    recursive=True,
    extension=".parquet",
    max_workers=None,
):
    """zipline does bundle ingestion one ticker at a time."""
    paths = list(
        glob.glob(os.path.join(by_ticker_dir, f"*{extension}"))
    )
    for path in paths:
        if os.path.exists(path):
            print(f"Removing {path}")
            os.remove(path)
    aggs_pattern = f"**/*{extension}" if recursive else f"*{extension}"
    paths = list(
        glob.glob(os.path.join(aggs_dir, aggs_pattern), recursive=recursive)
    )
    if max_workers == 0:
        for path in paths:
            split_aggs_by_ticker(
                path, by_ticker_dir=by_ticker_dir, extension=extension
            )
    else:
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            executor.map(
                split_aggs_by_ticker,
                paths,
                [by_ticker_dir] * len(paths),
                [extension] * len(paths),
            )


if __name__ == "__main__":
    os.environ["POLYGON_DATA_DIR"] = "/Volumes/Oahu/Mirror/files.polygon.io"
    config = PolygonConfig(environ=os.environ, calendar_name="XNYS", start_session=None, end_session=None)
    split_all_aggs_by_ticker(config.minute_aggs_dir)



from .config import PolygonConfig

from typing import Iterator

import argparse
import glob
import os

import pyarrow as pa
from pyarrow import dataset as pa_ds
from pyarrow import csv as pa_csv

PARTITION_COLUMN_NAME = "part"


# To work across all reasonable filesystems, we need to escape the characters in partition keys that are treated weirdly in filenames.
def partition_key_escape(c: str) -> str:
    return ("^" + c.upper()) if c.islower() else ("%" + "%02X" % ord(c))


def to_partition_key(s: str) -> str:
    if s.isalnum() and s.isupper():
        return s
    return "".join(
        [f"{c if (c.isupper() or c.isdigit()) else partition_key_escape(c)}" for c in s]
    )


def read_csv_table(path, timestamp_type: pa.TimestampType, convert_options):
    table = pa.csv.read_csv(path, convert_options=convert_options)
    table = table.set_column(
        table.column_names.index("window_start"),
        "window_start",
        table.column("window_start").cast(timestamp_type),
    )
    return table


def csv_agg_scanner(
    paths: list, schema: pa.Schema, timestamp_type: pa.TimestampType
) -> Iterator[pa.RecordBatch]:
    for path in paths:
        convert_options = pa_csv.ConvertOptions(
            column_types=schema,
            strings_can_be_null=False,
            quoted_strings_can_be_null=False,
        )

        print(f"{path=}")
        table = read_csv_table(
            path=path, timestamp_type=timestamp_type, convert_options=convert_options
        )

        table = table.append_column(
            PARTITION_COLUMN_NAME,
            pa.array(
                [to_partition_key(ticker) for ticker in table.column("ticker").to_pylist()]
            ),
        )

        for batch in table.to_batches():
            yield batch


def concat_all_aggs_from_csv(
    config: PolygonConfig,
    aggs_pattern: str = "**/*.csv.gz",
) -> list:
    """zipline does bundle ingestion one ticker at a time."""

    # We sort by path because they have the year and month in the dir names and the date in the filename.
    paths = sorted(
        list(
            glob.glob(
                os.path.join(config.aggs_dir, aggs_pattern),
                recursive="**" in aggs_pattern,
            )
        )
    )

    # Polygon Aggregate flatfile timestamps are in nanoseconds (like trades), not milliseconds as the docs say.
    # I make the timestamp timezone-aware because that's how Unix timestamps work and it may help avoid mistakes.
    timestamp_type = pa.timestamp("ns", tz="UTC")

    # But we can't use the timestamp type in the schema here because it's not supported by the CSV reader.
    # So we'll use int64 and cast it after reading the CSV file.
    # https://github.com/apache/arrow/issues/44030

    # strptime(3) (used by CSV reader for timestamps in ConvertOptions.timestamp_parsers) supports Unix timestamps (%s) and milliseconds (%f) but not nanoseconds.
    # https://www.geeksforgeeks.org/how-to-use-strptime-with-milliseconds-in-python/
    # Actually that's the wrong strptime (it's Python's).  C++ strptime(3) doesn't even support %f.
    # https://github.com/apache/arrow/issues/39839#issuecomment-1915981816
    # Also I don't think you can use those in a format string without a separator.

    # Polygon price scale is 4 decimal places (i.e. hundredths of a penny), but we'll use 10 because we have precision to spare.
    price_type = pa.decimal128(precision=38, scale=10)

    polygon_aggs_schema = pa.schema(
        [
            pa.field("ticker", pa.string(), nullable=False),
            pa.field("volume", pa.int64(), nullable=False),
            pa.field("open", price_type, nullable=False),
            pa.field("close", price_type, nullable=False),
            pa.field("high", price_type, nullable=False),
            pa.field("low", price_type, nullable=False),
            pa.field("window_start", pa.int64(), nullable=False),
            pa.field("transactions", pa.int64(), nullable=False),
        ]
    )

    partitioned_schema = polygon_aggs_schema.append(
        pa.field(PARTITION_COLUMN_NAME, pa.string(), nullable=False)
    )
    agg_scanner = pa_ds.Scanner.from_batches(
        csv_agg_scanner(paths=paths, schema=polygon_aggs_schema, timestamp_type=timestamp_type),
        schema=partitioned_schema
    )

    by_ticker_base_dir = os.path.join(
        config.by_ticker_dir,
        f"{config.agg_time}_{config.start_timestamp.date().isoformat()}_{config.end_timestamp.date().isoformat()}.hive",
    )
    partition_by_ticker = pa.dataset.partitioning(
        pa.schema([(PARTITION_COLUMN_NAME, pa.string())]), flavor="hive"
    )
    pa_ds.write_dataset(
        agg_scanner,
        base_dir=by_ticker_base_dir,
        format="parquet",
        partitioning=partition_by_ticker,
        existing_data_behavior="overwrite_or_ignore",
        max_partitions=config.max_partitions,
        max_open_files=config.max_open_files,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--calendar_name", default="XNYS")

    parser.add_argument("--start_session", default="2014-06-16")
    parser.add_argument("--end_session", default="2024-09-06")
    # parser.add_argument("--start_session", default="2020-10-07")
    # parser.add_argument("--end_session", default="2020-10-15")
    # parser.add_argument("--aggs_pattern", default="2020/10/**/*.csv.gz")
    parser.add_argument("--aggs_pattern", default="**/*.csv.gz")

    parser.add_argument("--overwrite", action="store_true")

    # TODO: These defaults should be None but for dev convenience they are set for my local config.
    parser.add_argument("--agg_time", default="day")
    parser.add_argument("--data_dir", default="/Volumes/Oahu/Mirror/files.polygon.io")

    args = parser.parse_args()

    # Maybe the way to do this is to use the os.environ as the argparser defaults.
    environ = dict(os.environ.items())
    if args.data_dir:
        environ["POLYGON_DATA_DIR"] = args.data_dir
    if args.agg_time:
        environ["POLYGON_AGG_TIME"] = args.agg_time

    config = PolygonConfig(
        environ=os.environ,
        calendar_name=args.calendar_name,
        start_session=args.start_session,
        end_session=args.end_session,
    )

    concat_all_aggs_from_csv(
        config=config,
        aggs_pattern=args.aggs_pattern,
    )



from .config import PolygonConfig

import os
import glob
from concurrent.futures import ProcessPoolExecutor
import pandas as pd


def convert_timestamp(x):
    """
    Polygon timestamps are in nanoseconds, milliseconds, or seconds.
    We can decide automatically based on the size of the number because the only overlaps
    are during the first few months of 1970.  And zero is always the same in any case.
    """
    try:
        unix_time = int(x)
        return pd.to_datetime(
            unix_time,
            unit=(
                "ns"
                if unix_time > 100_000_000_000_000
                else "ms" if unix_time > 10_000_000_000 else "s"
            ),
        )
    except Exception as e:
        print(f"ERROR: Failed to convert '{x}': {e}")
        return pd.NaT


def convert_csv_to_parquet(path, extension, compression):
    print(path)
    try:
        bars_df = pd.read_csv(
            path,
            compression=compression,
            converters={"window_start": convert_timestamp},
            dtype={"ticker": "str"},
        )
        # bars_df["ticker"] = bars_df["ticker"].astype(str)
        # bars_df.info()
        # print(f"{bars_df["ticker"].str.len().max()=}")
        if len(bars_df) == 0:
            print(f"WARNING: Empty {path}")
            return
        # if len(bars_df) < 100000:
        #     print(f"WARNING: Short {path}")
        # Don't change the data.  We're just converting to Parquet to save time.
        # bars_df.set_index(["window_start", "ticker"], inplace=True)
        # bars_df.sort_index(inplace=True)
        parquet_path = path.removesuffix(extension) + ".parquet"
        bars_df.to_parquet(parquet_path)
        # fp.write(parquet_path, bars_df, has_nulls=False, write_index=False, fixed_text={"ticker": bars_df["ticker"].str.len().max()})
        if not os.path.exists(parquet_path):
            print(f"ERROR: Failed to write {parquet_path}")
    except Exception as e:
        print(f"Failed for {path}: {e}")


def process_all_csv_to_parquet(
    aggs_dir,
    recursive=True,
    extension=".csv.gz",
    compression="infer",
    force=False,
    max_workers=None,
):
    """Big CSV files are very slow to read.  So we only read them once and convert them to Parquet."""
    csv_pattern = f"**/*{extension}" if recursive else f"*{extension}"
    paths = list(glob.glob(os.path.join(aggs_dir, csv_pattern), recursive=recursive))
    if force:
        print(f"Removing Parquet files that may exist for {len(paths)} CSV files.")
        for path in paths:
            parquet_path = path.removesuffix(extension) + ".parquet"
            if os.path.exists(parquet_path):
                print(f"Removing {parquet_path}")
                os.remove(parquet_path)
    else:
        csv_file_count = len(paths)
        paths = [
            path
            for path in paths
            if not os.path.exists(path.removesuffix(extension) + ".parquet")
        ]
        if len(paths) < csv_file_count:
            print(f"Skipping {csv_file_count - len(paths)} already converted files.")
    if max_workers == 0:
        for path in paths:
            convert_csv_to_parquet(path, extension=extension, compression=compression)
    else:
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            executor.map(
                convert_csv_to_parquet,
                paths,
                [extension] * len(paths),
                [compression] * len(paths),
            )


if __name__ == "__main__":
    # os.environ["POLYGON_DATA_DIR"] = "/Volumes/Oahu/Mirror/files.polygon.io"
    config = PolygonConfig(
        environ=os.environ, calendar_name="XNYS", start_session=None, end_session=None
    )
    process_all_csv_to_parquet(config.aggs_dir)



from exchange_calendars.calendar_helpers import Date, parse_date, parse_timestamp
from zipline.utils.calendar_utils import get_calendar

import os
import pandas as pd


class PolygonConfig:
    def __init__(
        self,
        environ: dict,
        calendar_name: str,
        start_session: Date,
        end_session: Date,
        agg_time: str = "day",
    ):
        if agg_time not in ["minute", "day"]:
            raise ValueError(f"agg_time must be 'minute' or 'day', got '{agg_time}'")
        self.calendar_name = calendar_name
        self.start_timestamp = (
            parse_date(start_session, calendar=self.calendar)
            if start_session
            else self.calendar.first_session
        )
        self.end_timestamp = (
            parse_date(end_session, calendar=self.calendar)
            if end_session
            else self.calendar.last_session
        )
        self.max_workers = None
        if environ.get("POLYGON_MAX_WORKERS", "").strip() != "":
            self.max_workers = int(environ.get("POLYGON_MAX_WORKERS"))
        self.api_key = environ.get("POLYGON_API_KEY")
        self.data_dir = environ.get("POLYGON_DATA_DIR", "data/files.polygon.io")
        self.cik_cusip_mapping_csv_path = environ.get(
            "CIK_CUSIP_MAPS_CSV", os.path.join(self.data_dir, "cik-cusip-maps.csv")
        )
        self.asset_subdir = environ.get("POLYGON_ASSET_SUBDIR", "us_stocks_sip")
        self.market = environ.get("POLYGON_MARKET", "stocks")
        self.tickers_dir = environ.get(
            "POLYGON_TICKERS_DIR",
            os.path.join(os.path.join(self.data_dir, "tickers"), self.asset_subdir),
        )
        self.tickers_csv_path = environ.get(
            "POLYGON_TICKERS_CSV",
            os.path.join(
                self.tickers_dir,
                f"tickers_{self.start_timestamp.date().isoformat()}_{self.end_timestamp.date().isoformat()}.csv",
            ),
        )
        self.flat_files_dir = environ.get(
            "POLYGON_FLAT_FILES_DIR", os.path.join(self.data_dir, "flatfiles")
        )
        self.csv_paths_pattern = environ.get("POLYGON_FLAT_FILES_CSV_PATTERN", "**/*.csv.gz")
        self.agg_time = agg_time
        self.asset_files_dir = os.path.join(self.flat_files_dir, self.asset_subdir)
        self.minute_aggs_dir = os.path.join(self.asset_files_dir, "minute_aggs_v1")
        self.day_aggs_dir = os.path.join(self.asset_files_dir, "day_aggs_v1")
        self.aggs_dir = (
            self.minute_aggs_dir if self.agg_time == "minute" else self.day_aggs_dir
        )
        # TODO: The "by ticker" files are temporary/intermediate and should/could be in the zipline data dir.
        self.minute_by_ticker_dir = os.path.join(
            self.asset_files_dir, "minute_by_ticker_v1"
        )
        self.day_by_ticker_dir = os.path.join(self.asset_files_dir, "day_by_ticker_v1")
        self.by_ticker_dir = (
            self.minute_by_ticker_dir
            if self.agg_time == "minute"
            else self.day_by_ticker_dir
        )
        self.arrow_format = environ.get("POLYGON_ARROW_FORMAT", "parquet" if self.agg_time == "day" else "hive")
        # self.by_ticker_hive_dir = os.path.join(
        #     self.by_ticker_dir,
        #     f"{self.agg_time}_{self.start_timestamp.date().isoformat()}_{self.end_timestamp.date().isoformat()}.hive",
        # )
        self.cache_dir = os.path.join(self.asset_files_dir, "api_cache")

    @property
    def calendar(self):
        return get_calendar(self.calendar_name)

    def ticker_file_path(self, date: pd.Timestamp):
        ticker_year_dir = os.path.join(
            self.tickers_dir, f"tickers_{date.strftime('%Y')}"
        )
        os.makedirs(ticker_year_dir, exist_ok=True)
        return os.path.join(
            ticker_year_dir, f"tickers_{date.date().isoformat()}.parquet"
        )
    
    def file_path_to_name(self, path: str):
        return os.path.basename(path).removesuffix(".gz").removesuffix(".csv")

    def by_ticker_aggs_arrow_dir(self, first_path: str, last_path: str):
        return os.path.join(
            self.by_ticker_dir,
            f"{self.file_path_to_name(first_path)}_{self.file_path_to_name(last_path)}.arrow",
        )

    def api_cache_path(
        self, start_date: Date, end_date: Date, filename: str, extension=".parquet"
    ):
        start_str = parse_date(start_date, calendar=self.calendar).date().isoformat()
        end_str = parse_date(end_date, calendar=self.calendar).date().isoformat()
        return os.path.join(
            self.cache_dir, f"{start_str}_{end_str}/{filename}{extension}"
        )


if __name__ == "__main__":
    config = PolygonConfig(os.environ, "XNYS", "2003-10-01", "2023-01-01")
    print(config.__dict__)



from .config import PolygonConfig

import polygon

import datetime
import logging
import os
import pandas as pd
from urllib3 import HTTPResponse


def load_polygon_splits(
    config: PolygonConfig, first_start_end: datetime.date, last_end_date: datetime.date
) -> pd.DataFrame:
    # N.B. If the schema changes then the filename should change.  We're on v3 now.
    splits_path = config.api_cache_path(
        start_date=first_start_end, end_date=last_end_date, filename="list_splits"
    )
    expected_split_count = (last_end_date - first_start_end).days * 3
    if not os.path.exists(splits_path):
        client = polygon.RESTClient(api_key=config.api_key)
        splits = client.list_splits(
            limit=1000,
            execution_date_gte=first_start_end,
            execution_date_lt=last_end_date + datetime.timedelta(days=1),
        )
        if splits is HTTPResponse:
            raise ValueError(f"Polygon.list_splits bad HTTPResponse: {splits}")
        splits = pd.DataFrame(splits)
        print(f"Got {len(splits)=} from Polygon list_splits.")
        os.makedirs(os.path.dirname(splits_path), exist_ok=True)
        splits.to_parquet(splits_path)
        if len(splits) < expected_split_count:
            logging.warning(
                f"Only got {len(splits)=} from Polygon list_splits (expected {expected_split_count=}).  "
                "This is probably fine if your historical range is short."
            )
        # We will always load from the file to avoid any chance of weird errors.
    if os.path.exists(splits_path):
        splits = pd.read_parquet(splits_path)
        print(f"Loaded {len(splits)=} from {splits_path}")
        if len(splits) < expected_split_count:
            logging.warning(
                f"Only got {len(splits)=} from Polygon list_splits (expected {expected_split_count=}).  "
                "This is probably fine if your historical range is short."
            )
        return splits
    raise ValueError(f"Failed to load splits from {splits_path}")


def load_splits(
    config: PolygonConfig,
    first_start_end: datetime.date,
    last_end_date: datetime.date,
    ticker_to_sid: dict[str, int],
) -> pd.DataFrame:
    splits = load_polygon_splits(config, first_start_end, last_end_date)
    splits["sid"] = splits["ticker"].apply(lambda t: ticker_to_sid.get(t, pd.NA))
    splits.dropna(inplace=True)
    splits["sid"] = splits["sid"].astype("int64")
    splits["execution_date"] = pd.to_datetime(splits["execution_date"])
    splits.rename(columns={"execution_date": "effective_date"}, inplace=True)
    # Not only do we want a float for ratio but some to/from are not integers.
    splits["split_from"] = splits["split_from"].astype(float)
    splits["split_to"] = splits["split_to"].astype(float)
    splits["ratio"] = splits["split_from"] / splits["split_to"]
    splits.drop(columns=["ticker", "split_from", "split_to"], inplace=True)
    return splits


def load_polygon_dividends(
    config: PolygonConfig, first_start_date: datetime.date, last_end_date: datetime.date
) -> pd.DataFrame:
    # N.B. If the schema changes then the filename should change.  We're on v3 now.
    dividends_path = config.api_cache_path(
        start_date=first_start_date, end_date=last_end_date, filename="list_dividends"
    )
    if not os.path.exists(dividends_path):
        client = polygon.RESTClient(api_key=config.api_key)
        dividends = client.list_dividends(
            limit=1000,
            record_date_gte=first_start_date,
            pay_date_lt=last_end_date + datetime.timedelta(days=1),
        )
        if dividends is HTTPResponse:
            raise ValueError(f"Polygon.list_dividends bad HTTPResponse: {dividends}")
        dividends = pd.DataFrame(dividends)
        os.makedirs(os.path.dirname(dividends_path), exist_ok=True)
        dividends.to_parquet(dividends_path)
        print(f"Wrote {len(dividends)=} from Polygon list_dividends to {dividends_path=}")
        # if len(dividends) < 10000:
        #     logging.error(f"Only got {len(dividends)=} from Polygon list_dividends.")
    # We will always load from the file to avoid any chance of weird errors.
    if os.path.exists(dividends_path):
        dividends = pd.read_parquet(dividends_path)
        # print(f"Loaded {len(dividends)=} from {dividends_path}")
        # if len(dividends) < 10000:
        #     logging.error(f"Only found {len(dividends)=} at {dividends_path}")
        return dividends
    raise ValueError(f"Failed to load dividends from {dividends_path}")


def load_chunked_polygon_dividends(
    config: PolygonConfig, first_start_end: datetime.date, last_end_date: datetime.date
) -> pd.DataFrame:
    dividends_list = []
    next_start_end = first_start_end
    while next_start_end < last_end_date:
        # We want at most a month of dividends at a time.  They should end on the last day of the month.
        # So the next_end_date is the day before the first day of the next month.
        first_of_next_month = datetime.date(
            next_start_end.year + (next_start_end.month // 12),
            (next_start_end.month % 12) + 1,
            1,
        )
        next_end_date = first_of_next_month - datetime.timedelta(days=1)
        if next_end_date > last_end_date:
            next_end_date = last_end_date
        dividends_list.append(load_polygon_dividends(
            config, next_start_end, next_end_date
        ))
        next_start_end = next_end_date + datetime.timedelta(days=1)
    return pd.concat(dividends_list)


def load_dividends(
    config: PolygonConfig,
    first_start_end: datetime.date,
    last_end_date: datetime.date,
    ticker_to_sid: dict[str, int],
) -> pd.DataFrame:
    dividends = load_chunked_polygon_dividends(config, first_start_end, last_end_date)
    dividends["sid"] = dividends["ticker"].apply(lambda t: ticker_to_sid.get(t, pd.NA))
    dividends.dropna(how="any", inplace=True)
    dividends["sid"] = dividends["sid"].astype("int64")
    dividends["declaration_date"] = pd.to_datetime(dividends["declaration_date"])
    dividends["ex_dividend_date"] = pd.to_datetime(dividends["ex_dividend_date"])
    dividends["record_date"] = pd.to_datetime(dividends["record_date"])
    dividends["pay_date"] = pd.to_datetime(dividends["pay_date"])
    dividends.rename(
        columns={
            "cash_amount": "amount",
            "declaration_date": "declared_date",
            "ex_dividend_date": "ex_date",
        },
        inplace=True,
    )
    dividends.drop(
        columns=["ticker", "frequency", "currency", "dividend_type"], inplace=True
    )
    return dividends



from .config import PolygonConfig

import shutil
from typing import Iterator, Tuple

import argparse
import glob
import os

import pyarrow as pa
from pyarrow import dataset as pa_ds
from pyarrow import csv as pa_csv

import pandas as pd


PARTITION_COLUMN_NAME = "part"
PARTITION_KEY_LENGTH = 2


def to_partition_key(s: str) -> str:
    """
    Partition key is low cardinality and must be filesystem-safe.
    The reason for partitioning is to keep the parquet files from getting too big.
    10 years of minute aggs for US stocks is 83GB gzipped.  A single parquet would be 62GB on disk.
    Currently the first two characters so files stay under 1GB.  Weird characters are replaced with "A".
    """
    k = (s + "A")[0:PARTITION_KEY_LENGTH].upper()
    if k.isalpha():
        return k
    # Replace non-alpha characters with "A".
    k = "".join([c if c.isalpha() else "A" for c in k])
    return k


def generate_tables_from_csv_files(
    paths: list,
    schema: pa.Schema,
    start_timestamp: pd.Timestamp,
    limit_timestamp: pd.Timestamp,
) -> Iterator[pa.Table]:
    empty_table = schema.empty_table()
    # TODO: Find which column(s) need to be cast to int64 from the schema.
    empty_table = empty_table.set_column(
        empty_table.column_names.index("window_start"),
        "window_start",
        empty_table.column("window_start").cast(pa.int64()),
    )
    csv_schema = empty_table.schema

    tables_read_count = 0
    skipped_table_count = 0
    for path in paths:
        convert_options = pa_csv.ConvertOptions(
            column_types=csv_schema,
            strings_can_be_null=False,
            quoted_strings_can_be_null=False,
        )

        table = pa.csv.read_csv(path, convert_options=convert_options)
        tables_read_count += 1
        table = table.set_column(
            table.column_names.index("window_start"),
            "window_start",
            table.column("window_start").cast(schema.field("window_start").type),
        )
        if schema.field(PARTITION_COLUMN_NAME) is not None:
            table = table.append_column(
                PARTITION_COLUMN_NAME,
                pa.array(
                    [
                        to_partition_key(ticker)
                        for ticker in table.column("ticker").to_pylist()
                    ]
                ),
            )
        expr = (
            pa.compute.field("window_start")
            >= pa.scalar(start_timestamp, type=schema.field("window_start").type)
        ) & (
            pa.compute.field("window_start")
            < pa.scalar(
                limit_timestamp,
                type=schema.field("window_start").type,
            )
        )
        table = table.filter(expr)

        # TODO: Also check that these rows are within range for this file's date (not just the whole session).
        # And if we're doing that (figuring date for each file), we can just skip reading the file.
        # Might able to do a single comparison using compute.days_between.
        # https://arrow.apache.org/docs/python/generated/pyarrow.compute.days_between.html

        if table.num_rows == 0:
            skipped_table_count += 1
            continue

        yield table
    print(f"{tables_read_count=} {skipped_table_count=}")


def generate_csv_agg_tables(
    config: PolygonConfig,
) -> Tuple[pa.Schema, Iterator[pa.Table]]:
    """zipline does bundle ingestion one ticker at a time."""
    # We sort by path because they have the year and month in the dir names and the date in the filename.
    paths = sorted(
        list(
            glob.glob(
                os.path.join(config.aggs_dir, config.csv_paths_pattern),
                recursive="**" in config.csv_paths_pattern,
            )
        )
    )

    print(f"{len(paths)=}")
    if len(paths) > 0:
        print(f"{paths[0]=}")
        print(f"{paths[-1]=}")

    # Polygon Aggregate flatfile timestamps are in nanoseconds (like trades), not milliseconds as the docs say.
    # I make the timestamp timezone-aware because that's how Unix timestamps work and it may help avoid mistakes.
    timestamp_type = pa.timestamp("ns", tz="UTC")

    # But we can't use the timestamp type in the schema here because it's not supported by the CSV reader.
    # So we'll use int64 and cast it after reading the CSV file.
    # https://github.com/apache/arrow/issues/44030

    # strptime(3) (used by CSV reader for timestamps in ConvertOptions.timestamp_parsers) supports Unix timestamps (%s) and milliseconds (%f) but not nanoseconds.
    # https://www.geeksforgeeks.org/how-to-use-strptime-with-milliseconds-in-python/
    # Actually that's the wrong strptime (it's Python's).  C++ strptime(3) doesn't even support %f.
    # https://github.com/apache/arrow/issues/39839#issuecomment-1915981816
    # Also I don't think you can use those in a format string without a separator.

    # Polygon price scale is 4 decimal places (i.e. hundredths of a penny), but we'll use 10 because we have precision to spare.
    # price_type = pa.decimal128(precision=38, scale=10)
    # 64bit float a little overkill but avoids any plausible truncation error.
    price_type = pa.float64()

    polygon_aggs_schema = pa.schema(
        [
            pa.field("ticker", pa.string(), nullable=False),
            pa.field("volume", pa.int64(), nullable=False),
            pa.field("open", price_type, nullable=False),
            pa.field("close", price_type, nullable=False),
            pa.field("high", price_type, nullable=False),
            pa.field("low", price_type, nullable=False),
            pa.field("window_start", timestamp_type, nullable=False),
            pa.field("transactions", pa.int64(), nullable=False),
        ]
    )
    if config.agg_time == "minute":
        polygon_aggs_schema = polygon_aggs_schema.append(
            pa.field(PARTITION_COLUMN_NAME, pa.string(), nullable=False)
        )

    return (
        paths,
        polygon_aggs_schema,
        generate_tables_from_csv_files(
            paths=paths,
            schema=polygon_aggs_schema,
            start_timestamp=config.start_timestamp,
            limit_timestamp=config.end_timestamp + pd.to_timedelta(1, unit="day"),
        ),
    )


def generate_batches_from_tables(tables):
    for table in tables:
        for batch in table.to_batches():
            yield batch


def concat_all_aggs_from_csv(
    config: PolygonConfig,
    overwrite: bool = False,
) -> str:
    paths, schema, tables = generate_csv_agg_tables(config)

    if len(paths) < 1:
        raise ValueError(f"No Polygon CSV flat files found in {config.aggs_dir=}")
    by_ticker_aggs_arrow_dir = config.by_ticker_aggs_arrow_dir(paths[0], paths[-1])
    if os.path.exists(by_ticker_aggs_arrow_dir):
        if overwrite:
            print(f"Removing {by_ticker_aggs_arrow_dir=}")
            shutil.rmtree(by_ticker_aggs_arrow_dir)
        else:
            print(f"Found existing {by_ticker_aggs_arrow_dir=}")
            return by_ticker_aggs_arrow_dir

    partitioning = None
    if schema.field(PARTITION_COLUMN_NAME) is not None:
        partitioning = pa_ds.partitioning(
            pa.schema([(PARTITION_COLUMN_NAME, pa.string())]), flavor="hive"
        )

    # scanner = pa_ds.Scanner.from_batches(source=generate_batches_from_tables(tables), schema=schema)
    pa_ds.write_dataset(
        generate_batches_from_tables(tables),
        schema=schema,
        base_dir=by_ticker_aggs_arrow_dir,
        partitioning=partitioning,
        format="parquet",
        existing_data_behavior="overwrite_or_ignore",
    )
    print(f"Concatenated aggregates to {by_ticker_aggs_arrow_dir=}")
    return by_ticker_aggs_arrow_dir


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--calendar_name", default="XNYS")

    parser.add_argument("--start_session", default="2014-06-16")
    parser.add_argument("--end_session", default="2024-09-06")
    # parser.add_argument("--start_session", default="2020-01-01")
    # parser.add_argument("--end_session", default="2020-12-31")

    parser.add_argument("--agg_time", default="day")

    parser.add_argument("--overwrite", action="store_true")

    # TODO: These defaults should be None but for dev convenience they are set for my local config.
    parser.add_argument("--data_dir", default="/Volumes/Oahu/Mirror/files.polygon.io")
    # parser.add_argument("--aggs_pattern", default="**/*.csv.gz")
    # parser.add_argument("--aggs_pattern", default="2020/10/**/*.csv.gz")

    args = parser.parse_args()

    # Maybe the way to do this is to use the os.environ as the argparser defaults.
    if args.data_dir:
        os.environ["POLYGON_DATA_DIR"] = args.data_dir

    config = PolygonConfig(
        environ=os.environ,
        calendar_name=args.calendar_name,
        start_session=args.start_session,
        end_session=args.end_session,
        agg_time=args.agg_time,
    )

    concat_all_aggs_from_csv(
        config=config,
        overwrite=args.overwrite,
    )



from .config import PolygonConfig

import os
import glob
import pandas as pd
import fastparquet as fp
from pathlib import Path


def apply_to_aggs(
    aggs_path: Path,
    func: callable,
    config: PolygonConfig,
    start_timestamp: pd.Timestamp = None,
    end_timestamp: pd.Timestamp = None,
):
    aggs_pf = fp.ParquetFile(aggs_path)
    df = aggs_pf.to_pandas()
    # Drop rows with window_start not in the range config.start_session to config.end_session
    # Those are dates (i.e. time 00:00:00) and end_session is inclusive.
    if start_timestamp is None and end_timestamp is None:
        return func(df, aggs_path, config)
    if start_timestamp is None:
        start_timestamp = pd.Timestamp.min
    if end_timestamp is None:
        end_timestamp = pd.Timestamp.max
    else:
        end_timestamp = end_timestamp + pd.Timedelta(days=1)
    df = df[
        df["window_start"].between(start_timestamp, end_timestamp, inclusive="left")
    ]
    return func(df, aggs_path, config)


# TODO: Return iterable of results
def apply_to_all_aggs(
    aggs_dir: str,
    func: callable,
    config: PolygonConfig,
    start_timestamp: pd.Timestamp = None,
    end_timestamp: pd.Timestamp = None,
    aggs_pattern: str = "**/*.parquet",
    max_workers=None,
) -> list:
    """zipline does bundle ingestion one ticker at a time."""
    paths = list(
        glob.glob(os.path.join(aggs_dir, aggs_pattern), recursive="**" in aggs_pattern)
    )
    print(f"{len(paths)=}")
    print(f"{paths[:5]=}")
    if max_workers == 0:
        return [
            apply_to_aggs(
                path, func, config=config, start_timestamp=start_timestamp, end_timestamp=end_timestamp
            )
            for path in paths
        ]
    else:
        # with ProcessPoolExecutor(max_workers=max_workers) as executor:
        #     executor.map(
        #         aggs_max_ticker_len,
        #         paths,
        #     )
        print("Not implemented")
        return None


if __name__ == "__main__":
    # os.environ["POLYGON_DATA_DIR"] = "/Volumes/Oahu/Mirror/files.polygon.io"
    def max_ticker_len(df: pd.DataFrame, path: Path, config: PolygonConfig):
        print(f"{path=}")
        return 0 if df.empty else df["ticker"].str.len().max()

    config = PolygonConfig(
        environ=os.environ,
        calendar_name="XNYS",
        start_session="2020-10-07",
        end_session="2020-10-15",
    )
    print(f"{config.aggs_dir=}")
    max_ticker_lens = apply_to_all_aggs(
        config.aggs_dir,
        func=max_ticker_len,
        config=config,
        aggs_pattern="2020/10/**/*.parquet",
        start_timestamp=config.start_timestamp,
        end_timestamp=config.end_timestamp,
        max_workers=0,
    )
    print(f"{max_ticker_lens=}")
    print(f"{len(max_ticker_lens)=}")
    print(f"{max(max_ticker_lens)=}")

# 2016 to 2024
# len(max_ticker_lens)=2184
# max(max_ticker_lens)=10
# 2020-10-07 to 2020-10-15
# max_ticker_lens=[0, 0, 0, 0, 8, 8, 8, 8, 8, 8, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
# len(max_ticker_lens)=22
# max(max_ticker_lens)=8
